import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_array
from sklearn.base import TransformerMixin
from sklearn.decomposition import PCA
from copy import deepcopy
from sklearn.utils.validation import check_is_fitted


class _BaseEDR(TransformerMixin):
    """Base class for Effective Dimensionality Reduction.
    It performs a single step of dimensionality reduction.

    Warning: This class should not be used directly.
    Use derived classes instead.
    """

    def __init__(self, estimator, dr_transformer):
        self.estimator = estimator
        self.dr_transformer = dr_transformer
        self.components_ = None

    def _check_estimator_fitted(self):
        check_is_fitted(self.estimator, 'estimator_')

    def _check_dr_transformer(self):
        if not hasattr(self.dr_transformer, 'components_'):
            raise AttributeError('The transformer does not expose '
                                 '"components_" attribute')

    def fit(self, X, y=None, **opt_kws):
        if y is None:
            self._check_estimator_fitted()
        else:
            self.estimator.fit(X, y, **opt_kws)

        grad = self.estimator.predict_gradient(X)
        self.dr_transformer.fit(grad)
        self._check_dr_transformer()
        self._set_components_(self.dr_transformer.components_)
        return self

    def transform(self, X):
        check_is_fitted(self, 'components_')
        X = check_array(X)
        return np.dot(X, self.components_.T)

    def inverse_transform(self, X):
        check_is_fitted(self, 'components_')
        X = check_array(X)
        return np.dot(X, np.linalg.pinv(self.components_))

    def _set_components_(self, components):
        self.components_ = deepcopy(components)


class EffectiveDimensionalityReduction(_BaseEDR):

    def __init__(self, estimator, dr_transformer, normalize=True,
                 preprocess_by_pca=False, **pca_kwargs):
        self.normalize = normalize
        self.preprocess_by_pca = preprocess_by_pca
        self.pca_kwargs = pca_kwargs
        super(EffectiveDimensionalityReduction, self).__init__(
            estimator, dr_transformer)

    def fit(self, X, y=None, **opt_kws):
        X = self._fit_preprocessing(X)
        super(EffectiveDimensionalityReduction, self).fit(X, y, **opt_kws)
        return self

    def _fit_preprocessing(self, X, transform=True):
        if not self.normalize:
            if self.preprocess_by_pca:
                mes = 'To apply PCA prerpocessing, normalize should be True'
                raise ValueError(mes)
            return X
        sc = StandardScaler()
        X_preprocessed = sc.fit_transform(X)
        # initialize self.components_
        self.components_ = np.diag(1 / sc.scale_)
        # note that X will be centered during training to improve
        # robustness of GP models.
        # the transform step will be a pure linear map without a translation

        if self.preprocess_by_pca:
            pca = PCA(**self.pca_kwargs)
            X_preprocessed = pca.fit(X_preprocessed)
            # update self.components_
            self._set_components_(pca.components_)
        return X_preprocessed if transform else None

    def _set_components_(self, components):
        self.components_ = (
            deepcopy(components) if not hasattr(self, 'components_')
            else np.dot(components, self.components_))



from sklearn.base import RegressorMixin
from GPy.models import GPRegression
from .base import _BaseGP


class GaussianProcessRegressor(_BaseGP, RegressorMixin):

    def __init__(self, kernels=None, kernel_options=None, Y_metadata=None,
                 normalizer=None, noise_var=1.0, mean_function=None):
        self.normalizer = normalizer
        self.noise_var = noise_var

        super(GaussianProcessRegressor, self).__init__(
            kernels, kernel_options, Y_metadata, mean_function)

    def _get_model(self, X, y, kernel):
        return GPRegression(X, y, kernel, self.Y_metadata, self.normalizer,
                            self.noise_var, self.mean_function)



from sklearn.base import ClassifierMixin
from GPy.models import GPClassification
from .base import _BaseGP


class GaussianProcessClassifier(_BaseGP, ClassifierMixin):

    def _get_model(self, X, y, kernel):
        return GPClassification(X, y, kernel, self.Y_metadata,
                                self.mean_function)



import numpy as np
from sklearn.utils import check_X_y, assert_all_finite, check_array
from sklearn.utils.multiclass import check_classification_targets
from sklearn.base import BaseEstimator
from copy import deepcopy
from GPy import kern as gpy_kern
from abc import ABCMeta, abstractmethod
import six


class _BaseGP(six.with_metaclass(ABCMeta, BaseEstimator)):
    def __init__(self, kernels=None, kernel_options=None, Y_metadata=None,
                 mean_function=None):
        self.kernels = kernels
        self.kernel_options = kernel_options
        self.Y_metadata = Y_metadata
        self.mean_function = mean_function
        self.estimator_ = None

    def fit(self, X, y, **opt_kws):
        X, y = self._check_input(X, y)

        kernel = self._make_kernel(X)
        self.estimator_ = self._get_model(X, y, kernel)

        opt_kws.setdefault('messages', False)
        opt_kws.setdefault('max_iters', 1000)
        self.estimator_.optimize(**opt_kws)
        return self

    def _check_input(self, X, y):
        X, y = check_X_y(X, y, accept_sparse=False)
        if self._estimator_type == 'classifier':
            check_classification_targets(y)
        y = y[:, np.newaxis]
        return X, y

    def _make_kernel(self, X):
        # kernel will be initiated as 'rbf' in model automatically
        if self.kernels is None:
            return self.kernels

        if isinstance(self.kernels, str):
            self.kernels = [self.kernel]

        kernels = [getattr(gpy_kern, kern) for kern in self.kernels]
        input_dim = {'input_dim': X.shape[1]}

        if self.kernel_options is None:
            options = [input_dim] * len(self.kernels)
        elif len(kernels) == len(self.kernel_options):
            options = deepcopy(self.kernel_options)
            for opt in options:
                opt.update(input_dim)
        else:
            raise(ValueError)

        kernel = np.sum([kern(**opt) for kern, opt in zip(kernels, options)])
        return kernel

    @abstractmethod
    def _get_model(self, X, y, kenrel):
        pass

    def predict(self, X):
        X = check_array(X, accept_sparse=False)
        y_pred = self.estimator_.predict(X)[0][:, 0]
        assert_all_finite(y_pred)
        return y_pred

    def predict_variance(self, X):
        return self.estimator_.predict(X)[1]

    def predict_gradient(self, X):
        return self.estimator_.predictive_gradients(X)[0][:, :, 0]


